#!/usr/bin/env python3

from FFDHud import FFDHud
import sys
from PIL import Image
import time
import threading
import math

score = 0
game_over = False

class Entity:
	def __init__(self, sprite_path, pos = (0,0)):
		self.sprite = Image.open(sprite_path)
		self.pos = [float(pos[0]), float(pos[1])]
		self.blendMode = "replace"
	
	def __del__(self):
		self.sprite.close()

	def move_abs(self, pos : (float, float)):
		self.pos = [float(pos[0]), float(pos[1])]

	def move_rel(self, delta : (float, float)):
		self.pos[0] += float(delta[0])
		self.pos[1] += float(delta[1])

	def draw(self, hud):
		hud.graphics_buffer_blit(self.sprite,
								 (int(self.pos[0] - self.sprite.size[0] / 2), int(self.pos[1] - self.sprite.size[1] / 2)),
								 self.sprite.size,
								 self.blendMode)
	
	def check_hit(self, ball):
		return False
	
	def update(self, timedelta):
		pass

	def handle_input(self, hud : FFDHud):
		pass

	def top_edge(self):
		return self.pos[1] - self.sprite.size[1] / 2
	def bottom_edge(self):
		return self.pos[1] + self.sprite.size[1] / 2
	def left_edge(self):
		return self.pos[0] - self.sprite.size[0] / 2
	def right_edge(self):
		return self.pos[0] + self.sprite.size[0] / 2


class Player(Entity):
	def __init__(self, sprite_path : str, pos : (float, float), hud : FFDHud):
		Entity.__init__(self, sprite_path, pos)
		self.running = True
		self.input_thread = threading.Thread(target=Player.handle_input, args=(self, hud))
		self.input_thread.start()
		self.vel = [0.0,0.0]
		self.setpoint = [*self.pos]
		self.k_p = 0.25
	
	def __del__(self):
		self.running = False
		self.input_thread.join()
	
	def update(self, timedelta):
		error = self.setpoint[0] - self.pos[0]
		P = error * self.k_p
		self.vel[0] = P
		if self.vel[0] > 0 and self.right_edge() < 128 or self.vel[0] < 0 and self.left_edge() >= 0:
			self.pos[0] += self.vel[0]

	def handle_input(self, hud : FFDHud):
		while self.running:
			event = hud.get_encoder_event(wait=False)
			if event:
				self.setpoint[0] += -5.0 * event[0]
				self.setpoint[0] = max(min(self.setpoint[0], float(128 - self.sprite.size[0]/2)), float(0 + self.sprite.size[0]/2))

class Ball(Entity):
	def __init__(self, sprite_path : str, pos : (float, float), vel : (float, float)):
		Entity.__init__(self, sprite_path, pos)
		self.vel = [*vel]
		self.blendMode = "add"
	
	def check_bounce(self, entity):
		if self == entity:
			return
		if (
			self.vel[1] > 0
			and self.bottom_edge() >= entity.top_edge()
			and self.bottom_edge() < entity.bottom_edge()
			and self.pos[0] >= entity.left_edge()
			and self.pos[0] <= entity.right_edge()
		):
			self.vel[1] = -self.vel[1]
			self.pos[1] -= 1
			if hasattr(entity, "vel"):
				self.vel[0] += entity.vel[0] * 5
		elif (
			self.vel[1] < 0
			and self.top_edge() <= entity.bottom_edge()
			and self.top_edge() > entity.top_edge()
			and self.pos[0] >= entity.left_edge()
			and self.pos[0] <= entity.right_edge()
		):
			self.vel[1] = -self.vel[1]
			self.pos[1] += 1
		if (
			self.vel[0] < 0
			and self.left_edge() <= entity.right_edge()
			and self.left_edge() > entity.left_edge()
			and self.pos[1] >= entity.top_edge()
			and self.pos[1] <= entity.bottom_edge()
		):
			self.vel[0] = -self.vel[0]
			self.pos[0] += 1
		elif (
			self.vel[0] > 0
			and self.right_edge() >= entity.left_edge()
			and self.right_edge() < entity.right_edge()
			and self.pos[1] >= entity.top_edge()
			and self.pos[1] <= entity.bottom_edge()
		):
			self.vel[0] = -self.vel[0]
			self.pos[0] -= 1

	def update(self, timedelta : float):
		global game_over
		if self.left_edge() < 0:
			self.vel[0] = abs(self.vel[0])
		elif self.right_edge() >= 128:
			self.vel[0] = -abs(self.vel[0])
		if self.bottom_edge() >= 64:
			game_over = True
		elif self.top_edge() < 0:
			self.vel[1] = abs(self.vel[1])
		self.pos[0] += self.vel[0] * timedelta
		self.pos[1] += self.vel[1] * timedelta

class Brick(Entity):
	def __init__(self, sprite_path : str, pos : (float, float)):
		Entity.__init__(self, sprite_path, pos)
	
	def check_hit(self, ball):
		return(
			ball.right_edge() >= self.left_edge()
			and ball.left_edge() <= self.right_edge()
			and ball.top_edge() <= self.bottom_edge()
			and ball.bottom_edge() >= self.top_edge()
		)
		

def main():
	global score, game_over
	score = 0
	game_over = False
	HUD = FFDHud("/dev/ttyACM1")
	HUD.clear()
	HUD.print("Brick Breaker", invert=True)
	time.sleep(3)
	HUD.discard_input()
	entities = []
	player = Player("brick_breaker_assets/paddle.png", (HUD.sizex // 2, HUD.sizey - 5), HUD)
	ball = Ball("brick_breaker_assets/ball.png", (round(HUD.sizex / 2), round(HUD.sizey * 2 / 3)), (15, 20))
	entities.append(player)
	entities.append(ball)
	for i in range(0, HUD.sizex, 17):
		for j in range(0, HUD.sizey // 3, 7):
			entities.append(Brick("brick_breaker_assets/brick.png", (i + 10,j + 3)))
	timestamp = time.time()
	while not game_over:
		HUD.graphics_buffer_clear()
		now = time.time()
		timedelta = now - timestamp
		to_delete = []
		for entity in entities:
			if entity.check_hit(ball):
				score += 1
				to_delete.append(entity)
			ball.check_bounce(entity)
			entity.update(timedelta)
			entity.draw(HUD)
		for entity in to_delete:
			entities.remove(entity)
		to_delete.clear()
		timestamp = now
		HUD.graphics_buffer_show()
		HUD.print(f"\033[8;1H{score}")
	HUD.clear()
	HUD.move_cursor_abs(1, 1)
	HUD.print("GAME OVER", invert=True)
	HUD.print(f"Score: {score}")
	time.sleep(3)
	player.__del__()


if __name__ == "__main__":
	main()
