#!/usr/bin/env python3

from FFDHud import FFDHud, FFDHudMenu
import subprocess, re, json, sys
from functools import partial

def run_command(hud : FFDHud, cmd : [str], shell : bool = False, confirm : bool = False):
	if confirm and not hud.confirm("Execute `" + " ".join(cmd) + "'?"):
		return
	try:
		hud.less(subprocess.check_output(cmd, shell=shell))
	except Exception as e:
		hud.less(str(e))

def show_lan_info(hud : FFDHud):
	interfaces = []
	try:
		for line in subprocess.check_output(["ip", "addr"]).decode(encoding="utf-8").split("\n"):
			match = re.search("^\s+inet\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/\d{1,2}).*\s(\w+)$", line)
			if match:
				interfaces.append(str(match.group(2)) + ": " + str(match.group(1)))
		hud.less("\n\n".join(interfaces))
	except Exception as e:
		hud.less(str(e))

def lsdev(hud : FFDHud):
	try:
		string = ""
		obj = json.loads(subprocess.check_output(["lsdev", "--json", "-dc"]))
		for row in obj["rows"]:
			for bay in row:
				string += bay["bay-id"].ljust(4) + " "
				if bay["occupied"]:
					string += bay["dev"] + " " + re.sub("\.\d{2}", "", bay["capacity"])
				else:
					string += "-"
				string += '\n'
		hud.less(string)
	except Exception as e:
		hud.less(str(e))

class SystemdUnit(FFDHudMenu):
	def __init__(self, hud : FFDHud, name : str, parent_menu : FFDHudMenu):
		super().__init__(hud, name)
		self.add_option("Status", self.status)
		self.start_option = self.add_option("Start", self.start)
		self.stop_option = self.add_option("Stop", self.stop)
		self.enable_option = self.add_option("Enable", self.enable)
		self.disable_option = self.add_option("Disable", self.disable)
		self.add_option("Back", self.hide)
		parent_menu.add_submenu(self)
	
	def is_enabled(self):
		try:
			string = subprocess.check_output(["systemctl", "is-enabled", self.name])
		except subprocess.CalledProcessError as e:
			string = e.output
		if string == b"enabled\n":
			return True
		elif string == b"disabled\n":
			return False
		return None

	def is_active(self):
		try:
			string = subprocess.check_output(["systemctl", "is-active", self.name])
		except subprocess.CalledProcessError as e:
			string = e.output
		if string == b"active\n":
			return True
		elif string == b"inactive\n":
			return False
		return None

	def status(self):
		output = self.formatted_name(self.hud.cols, scroll=False)
		status = subprocess.check_output(["systemctl status " + self.name + " --no-pager | perl -pne 's/[\t ]+/ /g;s/^[\t ]//g'"], shell=True).decode(encoding="utf-8")
		loaded_enabled = re.search("^Loaded: (\w+)[^;]+; (\w+)", status, re.MULTILINE)
		if loaded_enabled:
			output += "\n" + loaded_enabled.group(1) + "; " + loaded_enabled.group(2)
		active = re.search("^Active: (\w+\s\(\w+\))", status, re.MULTILINE)
		if active:
			output += "\n" + active.group(1)
		since = re.search("^Active: .*since ([^;]+); (.*)$", status, re.MULTILINE)
		if since:
			output += "\nsince\n"
			output += since.group(1) + "\n"
			output += "(" + since.group(2) + ")"
		trig = re.search("^TriggeredBy: . (.*)$", status, re.MULTILINE)
		if trig:
			output += "\nTrig: " + trig.group(1)
		self.hud.less(output)

	def start(self):
		if not self.hud.confirm("Start " + self.name + "?"):
			return
		self.hud.clear()
		self.hud.print("Starting " + self.name + "...", reflow=True)
		try:
			subprocess.run(["systemctl", "start", self.name])
		except subprocess.CalledProcessError as e:
			self.hud.less(e.output)
		except Exception as e:
			self.hud.less(str(e))
		self.update_options()
		self.hud.discard_input()
	
	def stop(self):
		if not self.hud.confirm("Stop " + self.name + "?"):
			return
		self.hud.clear()
		self.hud.print("Stopping " + self.name + "...", reflow=True)
		try:
			subprocess.run(["systemctl", "stop", self.name])
		except subprocess.CalledProcessError as e:
			self.hud.less(e.output)
		except Exception as e:
			self.hud.less(str(e))
		self.update_options()
		self.hud.discard_input()
	
	def enable(self):
		if not self.hud.confirm("Enable " + self.name + "?"):
			return
		self.hud.clear()
		self.hud.print("Enabling " + self.name + "...", reflow=True)
		try:
			subprocess.run(["systemctl", "enable", self.name])
		except subprocess.CalledProcessError as e:
			self.hud.less(e.output)
		except Exception as e:
			self.hud.less(str(e))
		self.update_options()
		self.hud.discard_input()
	
	def disable(self):
		if not self.hud.confirm("Disable " + self.name + "?"):
			return
		self.hud.clear()
		self.hud.print("Disabling " + self.name + "...", reflow=True)
		try:
			subprocess.run(["systemctl", "disable", self.name])
		except subprocess.CalledProcessError as e:
			self.hud.less(e.output)
		except Exception as e:
			self.hud.less(str(e))
		self.update_options()
		self.hud.discard_input()
	
	def update_options(self):
		enabled = self.is_enabled()
		if enabled:
			self.enable_option.hidden = True
			self.disable_option.hidden = False
		elif enabled == False:
			self.enable_option.hidden = False
			self.disable_option.hidden = True
		else:
			self.enable_option.hidden = True
			self.disable_option.hidden = True
		active = self.is_active()
		if active:
			self.start_option.hidden = True
			self.stop_option.hidden = False
		else:
			self.start_option.hidden = False
			self.stop_option.hidden = True

	def show(self):
		self.update_options()
		super().show()

def show_systemd_units(hud : FFDHud, settings : dict):
	unit_menu = FFDHudMenu(hud, "systemd")
	units = []
	unit_regex = None
	try:
		unit_regex = settings["visibility"]["admin"]["systemd"]["unit-regex"]
	except Exception as e:
		print("Failed to load systemd unit regex from settings:", file=sys.stderr)
		print(e, file=sys.stderr)
		print("Defaulting to unit-regex=\"/\S+\.(service|socket|timer).*/g\"", file=sys.stderr)
		unit_regex = "/\S+\.(service|socket|timer).*/g"
	unit_records = subprocess.check_output(["systemctl list-units --no-pager --no-legend --output=short-precise --plain --full --all | perl -ne 'while(" + unit_regex + "){print \"$&\n\" =~ s/[\t ]+/:/gr;}'"], shell=True).decode(encoding="utf-8").split("\n")
	unit_records.sort()
	for line in unit_records:
		if re.match("^\s*$", line):
			continue
		cols = line.split(":")
		name = cols[0]
		units.append(SystemdUnit(hud, name, unit_menu))
	unit_menu.add_option("Back", unit_menu.hide)
	unit_menu.show()

def determine_if_show(visibility, section, key):
	show_all_section = section not in visibility.keys()
	return show_all_section or visibility[section][key] if key in visibility[section].keys() else True

def main():
	if len(sys.argv) != 2:
		print("should only have one argument of device file name")
		sys.exit(1)
	device_path = "/dev/" + sys.argv[1]
	settings = {}
	try:
		with open("/etc/45D_HUD/menu-settings.json", "r") as f:
			settings = json.load(f)
	except Exception as e:
		print("Failed to load configuration file /etc/45D_HUD/menu-settings.json:", file=sys.stderr)
		print(e, file=sys.stderr)
	hud = FFDHud(device_path, idle=settings["idle-seconds"] if "idle-seconds" in settings.keys() else 30)
	main_menu = FFDHudMenu(hud)

	show_all = "visibility" not in settings.keys()

	show_all_info = show_all or "info" not in settings["visibility"].keys()
	if show_all_info or determine_if_show(settings["visibility"], "info", "self"):
		info_menu = FFDHudMenu(hud, "Info")
		if show_all_info or determine_if_show(settings["visibility"], "info", "temperature"):
			info_menu.add_option("Temperature", partial(run_command, hud, ["sensors | perl -pne 's/\s+\([^\)]+\)$//g;s/[\t ]+/ /g'"], shell=True))
		if show_all_info or determine_if_show(settings["visibility"], "info", "lan"):
			info_menu.add_option("LAN", partial(show_lan_info, hud))
		if show_all_info or determine_if_show(settings["visibility"], "info", "ipmi-lan"):
			info_menu.add_option("IPMI LAN", partial(
					run_command, hud,
					["sudo ipmitool lan print | perl -ne 'while(/(IP Address|Subnet Mask|MAC Address|Default Gateway IP).*/g){print \"$&\n\n\" =~ s/\s+:/:/gr;}'"],
					shell=True
				)
			)
		if show_all_info or determine_if_show(settings["visibility"], "info", "lsdev"):
			info_menu.add_option("lsdev", partial(lsdev, hud))
		info_menu.add_option("Back", info_menu.hide)
		main_menu.add_submenu(info_menu)

	show_all_admin = show_all or "admin" not in settings["visibility"].keys()
	if show_all_admin or determine_if_show(settings["visibility"], "admin", "self"):
		admin_menu = FFDHudMenu(hud, "Administration")
		if show_all_admin or determine_if_show(settings["visibility"], "admin", "dmap"):
			admin_menu.add_option("dmap", partial(run_command, hud, ["dmap"], confirm=True))
		if show_all_admin or determine_if_show(settings["visibility"], "admin", "dalias"):
			admin_menu.add_option("dalias", partial(run_command, hud, ["dalias"], confirm=True))
		if show_all_admin or determine_if_show(settings["visibility"], "admin", "systemd"):
			admin_menu.add_option("systemd", partial(show_systemd_units, hud, settings))
		admin_menu.add_option("Back", admin_menu.hide)
		main_menu.add_submenu(admin_menu)

	show_all_power = show_all or "power" not in settings["visibility"].keys()
	if show_all_power or determine_if_show(settings["visibility"], "power", "self"):
		power_menu = FFDHudMenu(hud, "Power")
		if show_all_power or determine_if_show(settings["visibility"], "power", "shutdown"):
			def shutdown(hud):
				if not hud.confirm("Run `shutdown -h now'?"):
					return
				hud.clear()
				hud.print("Shutting down...")
				subprocess.run(["shutdown", "-h", "now"])
			power_menu.add_option("Shutdown", partial(shutdown, hud))
		if show_all_power or determine_if_show(settings["visibility"], "power", "reboot"):
			def reboot(hud):
				if not hud.confirm("Run `reboot'?"):
					return
				hud.clear()
				hud.print("Rebooting...")
				subprocess.run(["reboot"])
			power_menu.add_option("Reboot", partial(reboot, hud))
		power_menu.add_option("Back", power_menu.hide)
		main_menu.add_submenu(power_menu)

	if not show_all and "games" in settings["visibility"].keys() and settings["visibility"]["games"] == True:
		try:
			import brick_breaker
			def start_brick_breaker(hud : FFDHud):
				hud.stop_idle_timer_thread()
				brick_breaker.main()
				hud.start_idle_timer_thread()
			game_menu = FFDHudMenu(hud, "Games")
			game_menu.add_option("Brick Breaker", partial(start_brick_breaker, hud))
			game_menu.add_option("Back", game_menu.hide)
			main_menu.add_submenu(game_menu)
		except Exception as e:
			print(e, file=sys.stderr)
			print("Failed to load games.", file=sys.stderr)

	main_menu.show()

if __name__ == "__main__":
	main()
