"""
	FFDHud - Library for interfacing with the 45Drives HUD

    Copyright (C) 2021 Joshua Boudreau <jboudreau@45drives.com>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import select
import tty
import textwrap
import time
import threading
from PIL import Image, ImageOps
from typing import Union, Literal, Tuple, List

def reflow_text_and_split(string : str, cols : int):
	return "\n".join(["\n".join([item.ljust(cols) for item in textwrap.wrap(output_row, width=cols)]) for output_row in string.split("\n")]).split("\n")

class FFDHud:
	"""
	Class to control a 45D HUD screen

	Constructor - opens character file in O_RDWR mode and sets tty to raw

	device_path : str
		Path to character device file for the 45D HUD
	sizex : int
		width of screen in pixels, default 128
	sizey : int
		height of screen in pixels, default 64
	idle : int
		If this many seconds pass without input, the splash screen will be displayed.
		If idle == 0, it will never timeout
	"""
	def __init__(self, device_path : str, sizex : int = 128, sizey : int = 64, idle : int = 0):
		self.fd = os.open(device_path, os.O_RDWR)
		tty.setraw(self.fd)
		self.graphics_buffer = bytearray(1024)
		self.sizex = sizex
		self.sizey = sizey
		self.rows = sizey // 8
		self.cols = sizex // 6
		self.mutex = threading.Lock()
		self.discard_input(timeout=1.0)
		self.left_arrow_char = b'\x9c'
		self.custom_char(self.left_arrow_char[0], bytearray([0x08, 0x1C, 0x2A, 0x08, 0x08, 0x00]))
		self.right_arrow_char = b'\x9d'
		self.custom_char(self.right_arrow_char[0], bytearray([0x08, 0x08, 0x2A, 0x1C, 0x08, 0x00]))
		self.up_arrow_char = b'\x9e'
		self.custom_char(self.up_arrow_char[0], bytearray([0x08, 0x04, 0x7E, 0x04, 0x08, 0x00]))
		self.down_arrow_char = b'\x9f'
		self.custom_char(self.down_arrow_char[0], bytearray([0x10, 0x20, 0x7E, 0x20, 0x10, 0x00]))
		if idle != 0:
			self.idle = idle
			self.start_idle_timer_thread()

	def __del__(self):
		self.stop_idle_timer_thread()
		os.close(self.fd)

	def idle_timer_thread_func(self):
		while self.idle_timer_thread_running:
			if self.idle_timer > 0:
				self.idle_timer -= 1
			elif self.idle_timer == 0:
				self.splash()
			time.sleep(1)
	
	def start_idle_timer_thread(self, idle : int = 0):
		"""
		Start the timer thread to show splash screen after `idle` seconds have passed
		with no input.

		idle : int
			Amount of seconds until idling, if 0 then no change from previous setting
		"""
		if idle > 0:
			self.idle = idle
		self.stop_idle_timer_thread()
		if hasattr(self, "idle"):
			self.idle_timer = self.idle
			self.idle_timer_thread_running = True
			self.idle_timer_thread = threading.Thread(target=self.idle_timer_thread_func)
			self.idle_timer_thread.start()
	
	def stop_idle_timer_thread(self):
		"""
		Stop the idle timer thread.
		"""
		if hasattr(self, "idle_timer_thread") and self.idle_timer_thread.is_alive():
			self.idle_timer_thread_running = False
			self.idle_timer_thread.join()

	def discard_input(self, timeout : int = 0.0) -> None:
		"""
		Throw out any buffered input events
		"""
		if select.select([self.fd],[],[],timeout)[0]:
			os.set_blocking(self.fd, False)
			while True:
				try:
					os.read(self.fd, 1024)
				except BlockingIOError:
					break
			os.set_blocking(self.fd, True)

	def write(self, array : Union[bytearray, bytes]) -> int:
		"""
		Write array of bytes to HUD, useful for graphics and escaped commands

		array : bytearray | bytes
			array of bytes to send to screen
		
		return : int
			number of bytes written
		"""
		if isinstance(array, bytearray) or isinstance(array, bytes):
			result = 0
			self.mutex.acquire()
			try:
				result = os.write(self.fd, array)
			finally:
				self.mutex.release()
			return result
		else:
			raise TypeError("Invalid type for FFDHud.write(): " + str(type(array)))
	
	def print(self, string : str = "", eol : Union[str, None] = '\n',
			  invert : Literal[True, False] = False,
			  underline : Literal[True, False] = False,
			  strike : Literal[True, False] = False,
			  reflow: Literal[True, False] = False) -> None:
		"""
		Print a string to the screen

		string : str
			The string to print
		eol : str | None
			End of line character, default '\\n'. Set to None to not send newline
		invert : True | False
			Invert text, default False
		underline : True | False
			underline text, default False
		strike : True | False
			Strikethrough text, defautl False
		reflow : True | False
			Reflow text to not break words
		"""
		if isinstance(string, str):
			if reflow:
				string = "\n".join(textwrap.wrap(string, width=self.cols))
			if eol:
				string += eol
			if invert:
				string = "\033[7m" + string
			if underline:
				string = "\033[4m" + string
			if strike:
				string = "\033[9m" + string
			if invert or underline or strike:
				string += "\033[0m"
			try:
				self.write(bytes(string, encoding="ascii"))
			except UnicodeEncodeError as e:
				stripped_string = ""
				for i in range(len(string)):
					character = string[i]
					if ord(character) > 0x9F:
						if character == '\xb0': # degrees
							stripped_string += chr(0x7f) # font table has degrees here
						else:
							stripped_string += '?'
					else:
						stripped_string += character
				try:
					self.write(bytes(stripped_string, encoding="ascii"))
				except:
					pass
			except:
				pass
		else:
			raise TypeError("Invalid type for FFDHud.print(): " + str(type(string)))
	
	def less(self, string : Union[str, bytes], show_return : Literal[True, False] = True):
		"""
		Display message with paging like less
		"""
		if isinstance(string, bytes):
			string = string.decode("utf-8")
		self.clear()
		self.discard_input()
		scroll = 0
		rows = reflow_text_and_split(string, self.cols)
		while True:
			first_row = scroll
			last_row = scroll + self.rows - 1
			self.print("\n".join([item.ljust(self.cols) for item in rows[first_row : min(last_row, len(rows))]]), reflow=False)
			if show_return:
				self.move_cursor_abs(self.rows, self.cols)
				self.write(b"\033[7m" + self.left_arrow_char + b"\033[0m")
			if last_row < len(rows):
				self.move_cursor_abs(self.rows, 1)
				self.print("More", invert=True, eol=None)
				self.print("  ")
			elif last_row == len(rows) and len(rows) != self.rows - 1:
				self.move_cursor_abs(self.rows, 1)
				self.print("Bottom", invert=True)
			event = self.get_encoder_event(wait=True)
			if event:
				if event[0]:
					scroll = max(min(scroll + event[0], len(rows) - self.rows + 1), 0)
				if event[1]:
					break

	def confirm(self, prompt : str, reflow : bool = True, affirmative : str = "Yes", negative : str = "No") -> bool:
		"""
		Display confirmation message and return user response.

		prompt : str
			the message to display
		reflow : bool
			whether or not to reflow the prompt
		affirmative : str
			String to show for option to return True
		negative : str
			String to show for option to return False
		"""
		choice = False
		affirmative = " " + affirmative + " "
		negative = " " + negative + " "
		self.clear()
		while True:
			self.move_cursor_abs(1, 1)
			self.print(prompt, reflow=reflow)
			self.move_cursor_abs(self.rows, 1)
			self.print(negative, invert=choice == False, eol = None)
			self.move_cursor_abs(self.rows, self.cols - len(affirmative) + 1)
			self.print(affirmative, invert=choice == True, eol = None)
			event = self.get_encoder_event(wait=True)
			if event:
				if event[1]:
					return choice
				if event[0] > 0: # CCW
					choice = False
				elif event[0] < 0: # CW
					choice = True

	
	def putc(self, char : str, eol : Union[str, None] = '\n',
			 invert : Literal[True, False] = False,
			 underline : Literal[True, False] = False,
			 strike : Literal[True, False] = False) -> None:
		"""
		Print single character without advancing cursor
		"""
		self.print(char, eol=None, invert=invert, underline=underline, strike=strike)
		self.move_cursor_rel(0, -1)

	def clear(self) -> None:
		"""
		Clear the screen
		"""
		self.write(b"\033C")
	
	def splash(self) -> None:
		"""
		Display splash screen
		"""
		self.write(b"\033S")

	def line_buffering(choice : Literal[True, False]) -> None:
		"""
		Set wether or not to line-buffer output internally in the screen
		"""
		if choice:
			self.write(b"\033BL")
		else:
			self.write(b"\033BC")
	
	def move_cursor_abs(self, row : int, col : int) -> None:
		"""
		Move text cursor to row,col with 1,1 being top left corner. (1-indexed because of ANSI escape codes)
		"""
		self.write(bytes(f"\033[{row};{col}H", encoding="ascii"))
	
	def move_cursor_rel(self, rowDelta : int, colDelta : int):
		"""
		Move text cursor relative to current position
		"""
		if rowDelta < 0:
			self.write(bytes(f"\033[{-rowDelta}A", encoding="ascii"))
		elif rowDelta > 0:
			self.write(bytes(f"\033[{rowDelta}B", encoding="ascii"))
		if colDelta < 0:
			self.write(bytes(f"\033[{-colDelta}D", encoding="ascii"))
		if colDelta > 0:
			self.write(bytes(f"\033[{colDelta}C", encoding="ascii"))

	def get_encoder_event(self, wait : Literal[True, False] = False) -> Union[Tuple[int, bool], None]:
		"""
		Get event of encoder state change

		wait : Boolean
			if True, will block until input, else will return None if no input ready

		return : tuple(int, Boolean) or None if !wait and no input ready
		"""
		if wait or select.select([self.fd],[],[],0.0)[0]:
			encoder_state = os.read(self.fd, 1)[0]
			if hasattr(self, "idle_timer"):
				if self.idle_timer == 0:
					self.clear()
					self.idle_timer = self.idle
					return None
				self.idle_timer = self.idle
			rotation = 0
			if (encoder_state & 0b001) and not (encoder_state & 0b010):
				rotation = -1
			elif (encoder_state & 0b010) and not (encoder_state & 0b001):
				rotation = 1
			press = (encoder_state & 0b100 != 0)
			return (rotation, press)
		return None
	
	def get_encoder_events(self) -> List[Tuple[int, bool]]:
		"""
		Get list of encoder events since last call

		return : [(int, bool)]
		"""
		events = []
		num_events = select.select([self.fd],[],[],0.0)[0]
		if num_events == 0:
			return events
		if hasattr(self, "idle_timer"):
			if self.idle_timer == 0:
				self.clear()
				self.idle_timer = self.idle
				return events
			self.idle_timer = self.idle
		for encoder_state in os.read(self.fd, num_events):
			rotation = 0
			if (encoder_state & 0b001) and not (encoder_state & 0b010):
				rotation = -1
			elif (encoder_state & 0b010) and not (encoder_state & 0b001):
				rotation = 1
			press = (encoder_state & 0b100 != 0)
			events.append((rotation, press))
		return events
	
	def graphics_buffer_set_pixel(self, x : int, y : int) -> None:
		"""
		Set a pixel in the graphics buffer

		x : int
			x position of pixel from left
		y : int
			y position of pixel from top
		"""
		if x >= self.sizex or x < 0 or y >= self.sizey or y < 0:
			return
		byte_slice = y % 8
		row_index = y // 8
		self.graphics_buffer[row_index * self.sizex + x] |= (1 << byte_slice)
	
	def graphics_buffer_clear_pixel(self, x : int, y : int) -> None:
		"""
		Cear a pixel in the graphics buffer

		x : int
			x position of pixel from left
		y : int
			y position of pixel from top
		"""
		if x >= self.sizex or x < 0 or y >= self.sizey or y < 0:
			return
		byte_slice = y % 8
		row_index = y // 8
		self.graphics_buffer[row_index * self.sizex + x] &= ~(1 << byte_slice)
	
	def graphics_buffer_clear(self) -> None:
		"""
		Clear the entire graphics buffer
		"""
		self.graphics_buffer = bytearray(1024)
	
	def graphics_buffer_show(self) -> None:
		"""
		Send graphics buffer to 45D HUD screen
		"""
		packet = bytearray(b"\033G")
		packet.extend(self.graphics_buffer)
		self.write(packet)
	
	def graphics_buffer_blit(self, image : Image.Image,
			pos : (int, int) = (0,0), size : (int, int) = (128,64),
			blendMode : Literal["replace", "lighten", "add", "darken", "subtract"] = "replace") -> None:
		"""
		Blit PIL Image to the graphics buffer

		image : PIL.Image.Image
			PIL image to blit
		pos : (int, int)
			Top left position of image in pixels
		size : (int, int)
			Width and height of image in pixels
		blendMode : "replace" | "lighten" | "add" | "darken" | "subtract"
			Method of blitting image
		"""
		if not isinstance(pos, tuple) or len(pos) != 2:
			raise TypeError("graphics_buffer_blit: pos must be 2D tuple")
		if not isinstance(size, tuple) or len(size) != 2:
			raise TypeError("graphics_buffer_blit: size must be 2D tuple")
		x = pos[0]
		y = pos[1]
		w = size[0]
		h = size[1]
		bitmap = image.resize((w,h)).convert(mode='1')
		if x < 0:
			bitmap = bitmap.crop((-x, 0, bitmap.size[0], bitmap.size[1]))
			x = 0
		if y < 0:
			bitmap = bitmap.crop((0, -y, bitmap.size[0], bitmap.size[1]))
			y = 0
		if x + w > self.sizex:
			bitmap = bitmap.crop((0, 0, x + self.sizex, bitmap.size[1]))
		if y + h > self.sizey:
			bitmap = bitmap.crop((0, 0, bitmap.size[0], y + self.sizey))
		method0 = None
		method1 = None
		if blendMode == "replace":
			method0 = self.graphics_buffer_clear_pixel
			method1 = self.graphics_buffer_set_pixel
		elif blendMode == "lighten" or blendMode == "add":
			method1 = self.graphics_buffer_set_pixel
		elif blendMode == "darken":
			method0 = self.graphics_buffer_clear_pixel
		elif blendMode == "subtract":
			method1 = self.graphics_buffer_clear_pixel
		else:
			raise "Invalid blend mode for blit"
		for i in range(bitmap.width):
			for j in range(bitmap.height):
				pixel = bitmap.getpixel((i,j))
				if pixel == 0 and method0 != None:
					method0(i + x, j + y)
				elif pixel != 0 and method1 != None:
					method1(i + x, j + y)

	def show_image(self, image : Image.Image,
				   fit : Literal["fill", "fit", "stretch"] = "fit") -> None:
		"""
		Simplest way to send an image to the screen.

		image : PIL.Image.Image
			The image to display
		fit : "fill" | "fit" | "stretch"
			Method of displaying.
			"fill" - fill entire screen maintaining aspect ratio
			"fit" - fit entire image maintaining aspect ratio
			"stretch" - resize image to screen dimensions, may warp
		"""
		self.graphics_buffer_clear()
		self.clear()
		x = None
		y = None
		w = None
		h = None
		image_aspect = image.size[0] / image.size[1]
		screen_aspect = self.sizex / self.sizey
		if fit == "fill":
			if image_aspect > screen_aspect:
				w = int(self.sizex * image_aspect / screen_aspect)
				x = int(self.sizex / 2 - w / 2)
				h = self.sizey
				y = 0
			else:
				w = self.sizex
				x = 0
				h = int(self.sizey * screen_aspect / image_aspect)
				y = int(self.sizey / 2 - h / 2)
		elif fit == "fit":
			if image_aspect > screen_aspect:
				w = self.sizex
				x = 0
				h = int(self.sizey * screen_aspect / image_aspect)
				y = int(self.sizey / 2 - h / 2)
			else:
				w = int(self.sizex * image_aspect / screen_aspect)
				x = int(self.sizex / 2 - w / 2)
				h = self.sizey
				y = 0
		else: # stretch
			x = 0
			y = 0
			w = self.sizex
			h = self.sizey
		self.graphics_buffer_blit(image, (x, y), (w, h), "replace")
		self.graphics_buffer_show()
	
	def custom_char(self, ordinal : Union[int, str], slices : bytearray):
		"""
		Add custom character to HUD

		ordinal : int | str
			0x80 <= ordinal <= 0x9F
		slices : bytearray
			array of 6 vertical slices representing character
		"""
		if isinstance(ordinal, str):
			ordinal = ord(ordinal)
		if ordinal < 0x80 or ordinal > 0x9F:
			raise ValueError("FFDHud.custom_char() requires 0x80 <= ordinal <= 0x9F")
		if len(slices) != 6:
			raise ValueError("FFDHud.custom_char() requires len(slices) == 6")
		command = bytearray("\033E", encoding="utf-8")
		command.append(ordinal)
		command += slices
		self.write(command)

class FFDHudMenuOption:
	"""
	An entry for an FFDHudMenu object. Use FFDHudMenu.add_option instead of constructing directly.

	name : str
		The text to display for the option
	action : callable
		function to run when option is chosen
	"""
	def __init__(self, name : str, action):
		if not callable(action):
			raise TypeError("FFDHudMenuOption requires callable action")
		self.name = name
		self.action = action
		self.hidden = False
	
	def formatted_name(self, cols : int, scroll : bool):
		"""
		Return the option's name, padded to the width of the screen. If the name is longer than the
		screen, the name is truncated with `..` in the middle.

		cols : int
			number of columns to pad/truncate to
		scroll : bool
			if true, long name will scroll
		"""
		name_len = len(self.name)
		if name_len < cols:
			return self.name.ljust(cols)
		if scroll:
			index = int(time.time() * 2) % (name_len - cols + 1)
			return self.name[index:index + cols]
		return (self.name[0 : cols // 2 - 1] + ".." + self.name[name_len - cols // 2 : ]).ljust(cols)

class FFDHudMenu(FFDHudMenuOption):
	"""
	Simple Menu System

	name : str
		Title displayed at the top of the screen, or if submenu, the entry name.
	"""
	def __init__(self, hud : FFDHud, name : str = "Main Menu"):
		FFDHudMenuOption.__init__(self, name, self.show)
		self.hud = hud
		self.options = []
		self.showing = False
		self.choice = 0
	
	def add_option(self, name : str, action) -> FFDHudMenuOption:
		"""
		Add an entry (FFDHudMenuEntry) to the menu.

		name : str
			The text to display for the option
		action : callable
			function to run when option is chosen
		"""
		option = FFDHudMenuOption(name, action)
		self.options.append(option)
		return option
	
	def add_submenu(self, submenu):
		"""
		Add another FFDHudMenu as a submenu. When chosen, the submenu's options
		will be displayed.
		"""
		if not isinstance(submenu, FFDHudMenu):
			raise TypeError("FFDHudMenu.add_submenu requires FFDHudMenu")
		self.options.append(submenu)

	def show(self):
		"""
		Display all options and wait for one to be chosen.
		"""
		self.showing = True
		self.hud.discard_input()
		visible_rows = self.hud.rows - 1
		self.hud.clear()
		options = list(filter(lambda opt: not opt.hidden, self.options))
		self.view_range = range(min(len(options), visible_rows))
		while self.showing:
			options = list(filter(lambda opt: not opt.hidden, self.options))
			self.hud.move_cursor_abs(1, 1)
			self.hud.print(self.formatted_name(self.hud.cols, scroll=True), underline=True)
			if self.choice > list(self.view_range)[-1]:
				self.view_range = range(self.choice - visible_rows + 1, self.choice + 1)
			elif self.choice < list(self.view_range)[0]:
				self.view_range = range(self.choice, self.choice + visible_rows)
			for i in self.view_range:
				selected = (i == self.choice)
				self.hud.print(options[i].formatted_name(self.hud.cols, scroll=selected), invert=selected)
			if len(options) > list(self.view_range)[-1] + 1:
				self.hud.move_cursor_abs(self.hud.rows, self.hud.cols)
				invert = self.choice == list(self.view_range)[-1]
				if invert:
					self.hud.write(b"\033[7m")
				self.hud.write(self.hud.down_arrow_char)
				if invert:
					self.hud.write(b"\033[0m")
			if list(self.view_range)[0] > 0:
				self.hud.move_cursor_abs(2, self.hud.cols)
				invert = self.choice == list(self.view_range)[0]
				if invert:
					self.hud.write(b"\033[7m")
				self.hud.write(self.hud.up_arrow_char)
				if invert:
					self.hud.write(b"\033[0m")
			encoder_event = self.hud.get_encoder_event(wait=(self.hud.idle_timer == 0 if hasattr(self.hud, "idle_timer") else False))
			if encoder_event:
				if encoder_event[0]:
					self.choice = max(min(self.choice + encoder_event[0], len(options) - 1), 0)
				elif encoder_event[1]:
					options[self.choice].action()
					self.hud.clear()
		self.hud.clear()
	
	def hide(self):
		"""
		Hide the menu, releasing it's control over the screen.
		"""
		self.showing = False
		
